//import 'package:flutter/foundation.dart';
import 'package:flutter/material.dart';
import 'package:flutter_test/flutter_test.dart';

typedef RouteBuilder = Widget Function(
    BuildContext context, Map<String, dynamic> params);

final Map<String, RouteBuilder> routes = {
  '/': (BuildContext context, Map<String, dynamic> params) =>
      const MyWidget(url: '/'),
  '/404': (context, params) => const MyWidget(url: '/404'),
  '/book/:iid': (context, params) => MyWidget(url: 'book/:iid=${params['id']}'),
  '/book/:iid/:iid1': (context, params) =>
      MyWidget(url: '/book,id=${params['id']},id1=${params['id1']}'),
};

class MyWidget extends StatelessWidget {
  final String url;
  const MyWidget({
    Key? key,
    required this.url,
  }) : super(key: key);

  @override
  Widget build(BuildContext context) {
    return MaterialApp(
      title: 'Flutter Demo',
      home: Scaffold(
        body: Center(
          child: Text(url),
        ),
      ),
    );
  }
}

void main() {
  test('should remove the unique id at the end of url', () {
    var url = '/_book_3';
    var result = UrlParser.removeOrder(url);
    expect(result, '/_book'); //_book should not matched

    RegExp reg = RegExp(r'(_[^/^_]+)$');
    expect(reg.hasMatch(url), true); //url中存在匹配reg的，最后的_3
    expect(reg.firstMatch(url)!.group(0), '_3'); //第一个匹配，group(0)，就是结果
    expect(reg.firstMatch(url)!.start, 6); //匹配序号为6
    expect(reg.allMatches(url).length, 1); //只有一个匹配

    //这个正则表达式应该安全，它必须是在尾部，且以_开头后面没有/
    //r'(_[^/^_]+)$'
    //$表示包括结束，也就是匹配在字符串最尾部
    //_表示以_开头
    //[^/^_]+ 表示不能有/和_两个字符，+表示多个字符。所以/_book_3只有一个匹配_3,book前的_到结尾之间还有_

    // var result = url.replaceAllMapped(r'(_[^/]+)$', (match) {
    //   print(
    //       'matched:${match.groupCount},group0:${match.group(0)},group1:${match.group(1)},group2:${match.group(2)}');
    //   return '';
    // });
  });

  test('get pattern from url', () {
    var url = '/book/:iid=1/:iid1=3:ipage=1';
    var result = UrlParser.getPattern(url);
    expect(result, '/book/:iid/:iid1');
  });

  test('get params from url', () {
    var url = '/book/:iid=1/:iid1=3:ipage=1';
    var result = UrlParser.getParam(url);
    expect(result['id'], 1);
    expect(result['id1'], 3);
    expect(result['page'], 1);
  });

  testWidgets('get widget from url', (WidgetTester tester) async {
    var url = '/';
    var pattern = UrlParser.getPattern(url);
    var params = UrlParser.getParam(url);
    var widget =
        Builder(builder: (context) => routes[pattern]!(context, params));
    await tester.pumpWidget(widget);
    expect(find.text('/'), findsOneWidget);
    expect(find.text('book'), findsNothing);
  });
}

class UrlParser {
  static const dynamicParams = {':ipage'};

  static String getPattern(url) {
    String pattern =
        url.replaceAllMapped(RegExp('(:[sid][^=]+)(=[^/]+)'), (match) {
      if (dynamicParams.contains(match.group(1))) return '';
      return '${match.group(1)}';
    });
    return pattern;
  }

  static Map<String, dynamic> getParam(url) {
    //support '/:iid=3/:iid1=2'
    //var regExp = RegExp('(/:[sid][^=]+)(=[^/]+)');

    //change to support‘/:iid=3:iid1=2'这样每节两个参数
    var regExp = RegExp('(:[sid][^=]+)(=[^/^:]+)');

    var result = regExp.allMatches(url);
    Map<String, dynamic> args = <String, dynamic>{};
    for (RegExpMatch match in result) {
      //only one param：
      //var argName = match.group(1)!.substring(3);
      //var argType = match.group(1)![2];

      var argName = match.group(1)!.substring(2);
      var argType = match.group(1)![1];

      var argValue = match.group(2)!.substring(1); //去掉等号

      dynamic value;

      if (argType == 'i') value = int.tryParse(argValue);
      if (argType == 'd') value = double.tryParse(argValue);
      if (argType == 's') value = argValue;
      args[argName] = value;
    }
    return args;
  }

  //like '/:iid=1_3这样，remove_3
  static removeOrder(String url) {
    var result = url.replaceAllMapped(RegExp(r'(_[^/^_]+)$'), (match) {
      return '';
    });
    return result;
  }
}
